/***************************************************************************
                          configuration.cpp  -  description
                             -------------------
    begin                : Tue Feb 1 2000
    copyright            : (C) 2000 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/


#include "configuration.h"
#include <map>
#include <string>
#include <set>
#include "fluff.h"

//=============================================================================
//This is a place to store the configuration and value of an option
class Config_value {
	public:
	Config_value(char *description_);
	~Config_value();

	const char *get_description() {return description;}

	const char *get_value() {return value.c_str();}
	void set_value(char *newvalue);
	void clear_value();

	//if true, then setting this value means appending
	//it to the old value, separated by a \t (tab) char.
	bool append_value;
	
	//true if this is an integer
	bool integer_value;

	//use for both strings and ints.
	int min;
	int max;

protected:
	char *description;
	std::string value;
};

//-----------------------------------------------------------------------------
Config_value::Config_value(char *description_)
{
	description = strdup(description_);
	value = "";
	append_value = false;
	integer_value = false;
	min = 0;
	max = 0;
}

//-----------------------------------------------------------------------------
Config_value::~Config_value()
{
	free(description);
}

//-----------------------------------------------------------------------------
void Config_value::clear_value()
{
	value = "";
}

//-----------------------------------------------------------------------------
void Config_value::set_value(char *newvalue)
{
	if (append_value) {

		if (value.size() > 0)
			value += "\t";
		
    value += newvalue;

	} else {
		value = newvalue;
	}
}

//=============================================================================
class Config_values : public std::map<std::string, Config_value *>{};

//=============================================================================
//-----------------------------------------------------------------------------
Configuration_observer::Configuration_observer()
{
	configuration = NULL;
}

//-----------------------------------------------------------------------------
Configuration_observer::~Configuration_observer()
{
	detatch();
}

//-----------------------------------------------------------------------------
void Configuration_observer::detatch()
{
	if (configuration) configuration->detach(this);
	configuration = NULL;
}

//-----------------------------------------------------------------------------
void Configuration_observer::config_attach(Configuration *config)
{
	detatch();
	configuration = config;
	if (configuration) configuration->attach(this);
}

//=============================================================================
class Config_observers : public std::set<Configuration_observer *>{};

//=============================================================================
//observer/subject stuff:
void Configuration::attach(Configuration_observer *observer)
{
	observers->insert(observer);	
}

//-----------------------------------------------------------------------------
void Configuration::detach(Configuration_observer *observer)
{
	observers->erase(observer);	
}

//-----------------------------------------------------------------------------
void Configuration::notify_observers()
{
  Config_observers::iterator i=observers->begin();
	while (i != observers->end()) {
    (*i)->config_changed();
  	i++;
	}
}

//=============================================================================
Configuration::Configuration()
{
	memset(error,0,sizeof(error));

	values = new Config_values;
	observers = new Config_observers;

	//set up all the options that we need to understand:
	add_option(
		"randomseed",
		"The random seed passed to the map generator (server only)",
		"42"
	);

	add_option(
		"data",
#ifdef _WIN32
		"The filename of either the .lump file or the directory where data can be found (default is: gfx/ripped and gfx/original)",
		"default.lump\tgfx/ripped\tgfx/original",
#else
		"The filename of either the .lump file or the directory where data can be found (default is: ~/nil/gfx/ripped and ~/nil/gfx/original)",
		"default.lump\t~/nil/gfx/ripped\t~/nil/gfx/original",
#endif
		0,0,false,
		true
	);

	add_option(
		"server_port",
		"The port to contact the server on or for the server to listen on (the default is 0xBABE)",
		"0xbabe",
		1,0xffff,true
	);

	add_option(
		"player_name",
		"The name presented to other players",
		"A. Coward"
	);

	add_option(
		"player_color",
		"The color of your worm",
		"0x80a0b0",
		0,0xffffff,true
	);

	add_option(
		"server",
		"The address of the server (the default is localhost)",
		"localhost"
	);

	add_option(
		"screen_size",
		"The size in pixels of the game screen (default is 640x480)",
		"640x480"
	);

	add_option(
		"world_size",
		"The size of the map in pixels.(server only) (default is 700x600)",
		"700x600"
	);

	add_option(
		"view_quality",
		"The number of game pixels pr screen pixel. 640x480 with quality 2 gives 320x240 in game resolution. (default is 2)",
		"2",
		1,5,true
	);

	add_option(
		"mode",
		"What to do: [server,client,clientserver,serverclient,cleanup,savelump] (default is clientserver)",
		"clientserver"
	);

	add_option(
		"output_lump",
		"What file to save the loaded data to, this is used by developers to create the .lump files",
		"default.lump"
	);

	add_option(
		"fullscreen",
		"Should the game run in fullscreen(1) or in a window (0) (default: 1)",
		"1",
		0,1,true
	);

	add_option(
		"coredump",
		"Should the game dump a core if something goes bad(1) or recover the system(0) (defaut:0)",
		"0",
		0,1,true
	);

	add_option(
		"net_multithread",
		"Should the server run an extra thread for each client(1) or run the network in a single thread(0) (default: 1)",
		"1",
		0,1,true
	);

	add_option(
		"net_normal_backlog",
		"The number of updates a client can be behind and still be considered uptodate, set to 5 for internet games (default: 1)",
		"1",
		0,20,true
	);

	add_option(
		"net_abnormal_backlog",
		"When a client gets this far behind it is kicked (default: 20)",
		"20",
		5,50,true
	);

	add_option(
		"net_max_wait_time",
		"The maximum time in ms to wait for all clients to become uptodate (default: 200)",
		"200",
		50,500,true
	);

	add_option(
		"chattiness",
		"The number of messages that NiL surts out. [5:assert, 4:debug, 3:message, 2:warning, 1:error, 0:fatal] (default: 4)",
		"4",
		0,5,true
	);


	//general purpose keys:
	add_option("key_Screen_shoot", 	    "The key bound to this function","-1",-1,255,true);
	add_option("key_Toggle_logo",       "The key bound to this function","-1",-1,255,true);
	add_option("key_Toggle_deathcam",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Toggle_traffic",    "The key bound to this function","-1",-1,255,true);
	add_option("key_Toggle_framerate",  "The key bound to this function","-1",-1,255,true);
	add_option("key_Toggle_shadow",     "The key bound to this function","-1",-1,255,true);
	add_option("key_Menu",              "The key bound to this function","-1",-1,255,true);

	//player 1 keys:
	add_option("key_Player1_up",        "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_down",      "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_left",      "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_right",     "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_fire",      "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_jump",      "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_dig",       "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_hook_fire", "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_hook_release","The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_hook_in",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_hook_out",  "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_weapon1",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_weapon2",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_weapon3",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_weapon4",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player1_weapon5",   "The key bound to this function","-1",-1,255,true);

	//player 2 keys
	add_option("key_Player2_up",        "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_down",      "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_left",      "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_right",     "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_fire",      "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_jump",      "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_dig",       "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_hook_fire", "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_hook_release","The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_hook_in",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_hook_out",  "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_weapon1",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_weapon2",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_weapon3",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_weapon4",   "The key bound to this function","-1",-1,255,true);
	add_option("key_Player2_weapon5",   "The key bound to this function","-1",-1,255,true);
}

//-----------------------------------------------------------------------------
#define CHECK_MIN(value,min)\
  if (value < min) {sprintf(error,"%s is %i this is less than the miniumum %s (%i)",#value,value,#min,min); return false;}
#define CHECK_MAX(value,max)\
  if (value > max) {sprintf(error,"%s is %i this is more than the maximum %s (%i)",#value,value,#max,max); return false;}

//-----------------------------------------------------------------------------
//This does secondary checks on the options:
bool Configuration::check_options()
{
	Config_xy screen_size = get_option_xy("screen_size");
	CHECK_MIN(screen_size.x,320)
	CHECK_MAX(screen_size.x,1280)
	CHECK_MIN(screen_size.y,200)
	CHECK_MAX(screen_size.y,1024)

	Config_xy world_size = get_option_xy("world_size");
	CHECK_MIN(world_size.x,640)
	CHECK_MAX(world_size.x,5000)
	CHECK_MIN(world_size.y,480)
	CHECK_MAX(world_size.y,5000)

	int view_quality = get_option_int("view_quality");
	CHECK_MIN(view_quality,(int)ceil(1.0*screen_size.x/world_size.x))
	CHECK_MIN(view_quality,(int)ceil(1.0*screen_size.y/world_size.y))

	const char *mode = get_option("mode");
	if (strcasecmp(mode,"client") &&	strcasecmp(mode,"server") &&
	    strcasecmp(mode,"clientserver") && strcasecmp(mode,"serverclient") &&
	    strcasecmp(mode,"cleanup") && strcasecmp(mode,"savelump")) {
  	sprintf(error,"The mode \"%s\" was not understood, valid modes are: server, client, clientserver, serverclient, savelump and cleeanup",mode);
    return false;
	}

	return true;
}


//-----------------------------------------------------------------------------
Configuration::~Configuration()
{
	delete observers;
	Config_values::iterator i=values->begin();
	while (i!=values->end()) {
    delete i->second;
		i++;
	}	
	delete values;
}

//-----------------------------------------------------------------------------
void Configuration::add_option(char *name, char *description, char *default_value, int min, int max, bool integer_value, bool append_value)
{
	Config_value *newvalue = new Config_value(description);
	
	newvalue->min = min;
	newvalue->max = max;
	newvalue->integer_value = integer_value;
	newvalue->append_value = append_value;
	newvalue->set_value(default_value);

	values->insert(std::pair<std::string, Config_value *>(name,newvalue));
}

//-----------------------------------------------------------------------------
bool Configuration::parse_cmdline(int argc,char **argv)
{
	//run through all argv's and parse_option() on them
	for (int c=1;c<argc;c++)
		if (!parse_option(argv[c])) return false;
		
	return true;
}

//-----------------------------------------------------------------------------
bool Configuration::load_file(char *filename)
{
	char expanded_filename[200];
	if (!expand_path(filename,expanded_filename,sizeof(expanded_filename))) {
  	sprintf(error,"Unable to expand the path %s",filename);
		return false;
	}

	FILE *f=fopen(expanded_filename,"r");
	if (!f) {
		sprintf(error,"Unable to open the configuration file %s",expanded_filename);
		return false;
	}

	fseek(f,0,SEEK_END);
	int length = ftell(f);
	fseek(f,0,SEEK_SET);
	
	char *file_contents = (char*)malloc(length+1);
  if ((length = fread(file_contents,1,length,f)) == 0) {
		sprintf(error,"Unable to read the configuration file %s",expanded_filename);
    free(file_contents);
  	fclose(f);
		return false;
	}
	fclose(f);
	file_contents[length] = 0; //terminate the file

	bool result = parse_file(file_contents);
	if (!result)
  	sprintf(error+strlen(error)," (while parsing %s)",expanded_filename);

  free(file_contents);

	return result;
}

//-----------------------------------------------------------------------------
bool Configuration::parse_file(char *file_contents)
{
	char *lines = strdup(file_contents);
	char *lines_saved = lines;
	
	char *line=strsep(&lines,"\n");
	while (line) {
		//eat leading whitepace:
		line = strtrim(line);

		//deal with the line if it is not blank and not a comment
		if (*line && *line != '#')
			if (!parse_option(line)) {
				free(lines_saved);
				return false;
			}

		line=strsep(&lines,"\n");
	}

	free(lines_saved);
	return true;
}

//-----------------------------------------------------------------------------
bool Configuration::parse_option(char *line)
{
	char *tmp = strdup(line);
	char *tmp_saved = tmp;

	//split on '=' and trim whitespace off of both the name and the value
	char *name = strsep(&tmp,"=");
	char *value = tmp;

	//note: values are allowed to be empty
	if (!name || !value || !*name) {
  	sprintf(error,"Unable to parse the option: \"%s\" options must all be on the form: name=value",line);
		free(tmp_saved);
		return false;
	}
	
	name = strtrim(name);
	value = strtrim(value);

	if (!*name) {
  	sprintf(error,"Unable to find a valid option name in the option: \"%s\" options must all be on the form: name=value",line);
		free(tmp_saved);
		return false;
	}

	bool result = set_option(name,value);
	
  free(tmp_saved);
	return result;
}

//-----------------------------------------------------------------------------
bool Configuration::clear_option(char *name)
{
	//find the option and set it to ""
	Config_values::iterator i=values->find(name);
	if (i == values->end()) {
  	sprintf(error,"Unable to find an option named %s to clear",name);
		return false;
	}

	i->second->clear_value();

	notify_observers();

	return true;
}

//-----------------------------------------------------------------------------
bool Configuration::set_option(char *name, char *value)
{
	if (!value || !*value) return clear_option(name);

	//check limits and set the value.
	Config_values::iterator i=values->find(name);
	if (i == values->end()) {
  	sprintf(error,"Unable to find an option named %s to set",name);
		return false;
	}

	if (i->second->integer_value) {

  	int ival = -1;
		if (sscanf(value,"%i",&ival) != 1) {
	  	sprintf(error,"Unable to set the option %s to %s (the value must be an integer)",name,value);
			return false;
		}
		
		if (ival < i->second->min) {
	  	sprintf(error,"Unable to set the option %s to %s (%i), the minimum is %i",name,value,ival,i->second->min);
			return false;
		}
			
		if (ival > i->second->max) {
	  	sprintf(error,"Unable to set the option %s to %s (%i), the maximum is %i",name,value,ival,i->second->max);
			return false;
		}
			
		//normalize the string so that it is always stored as decimal
		char tmp_value[40];
		sprintf(tmp_value,"%i",ival);
		
		i->second->set_value(tmp_value);

	} else {

    //string values:
		i->second->set_value(value);		

	}

	notify_observers();

	return true;
}

//-----------------------------------------------------------------------------
const char *Configuration::get_option(char *name)
{
	Config_values::iterator i=values->find(name);
	if (i == values->end()) {
  	sprintf(error,"Unable to find an option named %s to get",name);
		logmsg(lt_warning,error);
		return NULL;
	}

	return i->second->get_value();
}

//-----------------------------------------------------------------------------
int Configuration::get_option_int(char *name)
{
	const char *value = get_option(name);
	int result = -1;
	if (value)
		sscanf(value,"%i",&result);
	return result;
}

//-----------------------------------------------------------------------------
Config_xy Configuration::get_option_xy(char *name)
{
	const char *value = get_option(name);
	Config_xy result;
	result.x = result.y = -1;
	if (value)
		sscanf(value,"%ix%i",&result.x,&result.y);
	return result;
}

//-----------------------------------------------------------------------------
void Configuration::print_manual(FILE *target)
{
	//find the length of the longest option name:
	int name_width = 0;
	{
		Config_values::iterator i=values->begin();
		while (i != values->end()) {
			if (strncmp(i->first.c_str(),"key",3) != 0) {
				int len=i->first.size();
				if (len > name_width)	name_width = len;			
			}
			i++;
		}
	}

	//make up a format so the values will be neatly aligned:
	char format[500];
	sprintf(format,"%%%is : %%s",name_width);

	//actually print the stuff:
	{
		Config_values::iterator i=values->begin();
		while (i != values->end()) {

			if (strncmp(i->first.c_str(),"key",3) != 0) {
				fprintf(target,format,i->first.c_str(),i->second->get_description());
				if (i->second->integer_value)
					fprintf(target," Must be in the range (%i..%i)",i->second->min,i->second->max);
				fprintf(target,"\n");	
			}
	
			i++;
		}
	}
}

//-----------------------------------------------------------------------------
bool Configuration::save_options(char *filename)
{
	char expanded_filename[200];
	if (!expand_path(filename,expanded_filename,sizeof(expanded_filename))) {
  	sprintf(error,"Unable to expand the path %s",filename);
		return false;
	}

	FILE *f = fopen(expanded_filename,"w");
	if (!f) {
  	sprintf(error,"Unable to open %s for saving the options",expanded_filename);
		return false;
	}

	print_options(f);	
	fclose(f);

	return true;
}

//-----------------------------------------------------------------------------
//dumps a nicely commented configuration file
void Configuration::print_options(FILE *target)
{
	fprintf(target,"# NiL options:\n\n");
	Config_values::iterator i=values->begin();
	while (i != values->end()) {

		fprintf(target,"# %s\n",i->second->get_description());	
		if (i->second->integer_value)
			fprintf(target,"# Must be an integer in the range (%i..%i)\n",i->second->min,i->second->max);
		fprintf(target,"%s=%s\n\n",i->first.c_str(),i->second->get_value());	

		i++;
	}
}

