/***************************************************************************
                          serverworld.cpp  -  description
                             -------------------
    begin                : Wed Oct 20 1999
    copyright            : (C) 1999 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/


#include "systemheaders.h"
#include "serverworld.h"
#include "obj_base.h"
#include "all.h"
#include "netpackets.h"
#include "world_objects.h"
#include "soundsystem.h"

//-----------------------------------------------------------------------------------
void Serverworld::save_original_map()
{
	if (original_map) free(original_map);

	int bytes_to_move = map_buffer.get_xsize()*map_buffer.get_ysize()*sizeof(int32);
	original_map = (int32*)malloc(bytes_to_move);	
	memcpy(original_map,map_buffer.get_pixels_read(),bytes_to_move);	
}

//-----------------------------------------------------------------------------------
Serverworld::Serverworld(Loader *loader_, Soundsystem *soundsystem_) : World(loader_, soundsystem_)
{
	next_free_objectid = 1;

	Obj_scoreboard *sb = (Obj_scoreboard *)newobject(ot_scoreboard);
	if (sb)	scoreboard_id = sb->getid();

	original_map = NULL;

	killer_line = 0;
	last_reader_length = 0;
}

//-----------------------------------------------------------------------------------
Serverworld::~Serverworld()
{
	if (original_map) free(original_map);
}

//-----------------------------------------------------------------------------------
Obj_base *Serverworld::newobject(Object_type objecttype)
{
	Obj_base *newone = World::newobject(objecttype,  next_free_objectid++);

	return newone;
}
/*
//-----------------------------------------------------------------------------------
Player *Serverworld::newplayer();
{
	return World::newplayer(next_free_playerid++);
}
*/

//-----------------------------------------------------------------------------------
void Serverworld::server_think(Update_queue *update_queue)
{
	Serializer_reader *serializer_reader = update_queue->new_reader();
	serializer_reader->clear();

	{
		//Let the objects do their thinking (this collects thoughts)
		World_objects::iterator i = objects->begin();
		while (i != objects->end()) {
			if (!i->second->isdone())
				i->second->server_think();
 	
			i++;
		}
	}

	{
		//collect up all the changed states of the surviving objects to send to clients
		World_objects::iterator i = objects->begin();
		while (i != objects->end()) {
			if (i->second->isdirty()) {

				serializer_reader->wint(TT_OBJECT_STATE);
				serializer_reader->wint(i->second->getid());
				serializer_reader->wint(i->second->gettype());
				i->second->serialize(serializer_reader);
	
				i->second->beclean();
			}
	
			i++;
		}
	}

  //Tell the clients to let the dead objects have it.
	{
		World_objects::iterator i = objects->begin();
		while (i != objects->end()) {

			if (i->second->isdone()) {
				serializer_reader->wint(TT_OBJECT_DELETE);
				serializer_reader->wint(i->second->getid());
			}
	
			i++;
		}
	}

	//clean out the dead objects
	remove_dead_objects();

	kill_lonely_bits(serializer_reader);
	//send the time:
	serializer_reader->wint(TT_WORLD_STATE);
	serializer_reader->wfloat(gettime());
	
	//feed the stats
	last_reader_length = serializer_reader->getlength();

	//let the update_queue at the serialized data:
	update_queue->submit_update(serializer_reader);
}

//-----------------------------------------------------------------------------------
void Serverworld::makehole(int x, int y, int diameter)
{
	if (diameter > 16) diameter = 16;
	if (diameter < 1) diameter = 1;

	//do the actual hole putting
	holes[diameter-1]->draw_map(&map_buffer,&bgmap_buffer,x,y);
}

//-----------------------------------------------------------------------------------
void Serverworld::make_all_dirty()
{
	World_objects::iterator i = objects->begin();
	while (i != objects->end()) {
		i->second->setdirty();
 		i++;
	}
}

//-----------------------------------------------------------------------------------
bool Serverworld::handle_join_request(
		Net_joinrequest &joinrequest,			//input, comes from the client

		Obj_player_data *&new_player,  	//output, don't delete!
		Net_joinreply &joinreply,				//output
		char *&map_data                 //output, free() when done
	)
{
	new_player = (Obj_player_data *)newobject(ot_player_data);
	if (!new_player) return false;

	new_player->init(&joinrequest);
			
	joinreply.objectid = new_player->getid(); //id of your Obj_player_data, this should be passed to clientworld
	joinreply.map_xsize = map_buffer.get_xsize();
	joinreply.map_ysize = map_buffer.get_ysize();
	joinreply.randomseed = randomseed;

	unsigned int pixelcount = joinreply.map_xsize*joinreply.map_ysize;
	unsigned int pixelbytes = pixelcount*sizeof(int32);

	//Create the diff:
	int32 *diff = (int32*)malloc(pixelbytes);
	for (unsigned int i=0;i<pixelcount;i++) {
		int32 op = original_map[i];      				//original pixel
		int32 mp = map_buffer.get_pixels_read()[i]; 	//current pixel

		if (IS_PASSABLE(mp) && !IS_PASSABLE(op)) {
			diff[i] = PF_TYPE_HOLE; //invert the hole bit
		} else
			diff[i] = mp ^ op;
	}

	//compress the data:
	unsigned int buffer_size = ROUND(pixelbytes*1.002); //read zlib for the extra space.
	map_data = (char*)malloc(buffer_size);

	unsigned long compressed_size = pixelbytes;
	int zerror = compress(
			(unsigned char *)map_data,
			&compressed_size,
			(const unsigned char*)diff,
			pixelbytes
		);

	if (zerror == Z_OK) {
  	joinreply.zcompressed = true;
		joinreply.compressed_size = compressed_size;
		logmsg(lt_debug,"Compressed the diff to %i bytes",joinreply.compressed_size);
	} else {
	  memcpy(map_data,diff,pixelbytes);
  	joinreply.zcompressed = false;
		joinreply.compressed_size = pixelbytes;
		logmsg(lt_error,"Unable to compress the diff, the error was %i",zerror);
	}

	free(diff);

	return true;
}	

//-----------------------------------------------------------------------------
void Serverworld::kill_lonely_bits(Serializer_reader *serializer_reader)
{
//	logmsg(lt_error,"kill enable");

	int w = bgmap_buffer.get_xsize();
	const int32 *bgpixel = bgmap_buffer.get_pixels_read();
	int32 *fgpixel = map_buffer.get_pixels_write();
	bgpixel+=killer_line*w;
	fgpixel+=killer_line*w;
	for (int i=0;i<2;i++) {
		for (int x=0;x<w;x++) {
			if (
				IS_EARTH(*fgpixel) &&
				IS_PASSABLE(*(fgpixel-1)) +
				IS_PASSABLE(*(fgpixel+1)) +

				IS_PASSABLE(*(fgpixel-1-w)) +
				IS_PASSABLE(*(fgpixel  -w)) +
				IS_PASSABLE(*(fgpixel+1-w)) +

				IS_PASSABLE(*(fgpixel-1+w)) +
				IS_PASSABLE(*(fgpixel  +w)) +
				IS_PASSABLE(*(fgpixel+1+w)) >4) {
				//remove pixel here and tell the client about it.
				serializer_reader->wint(TT_DEAD_PIXEL);
				serializer_reader->wint(x);
				serializer_reader->wint(killer_line);
				*fgpixel=*bgpixel;				
				
				//logmsg(lt_error,"killed one!x,y:%d,%d",x,killer_line+i);
			}
			fgpixel++;
			bgpixel++;
		}

		killer_line++;
		if (killer_line >= bgmap_buffer.get_ysize()-1)
			killer_line = 1;
	}

}

void Serverworld::do_radius_damage(int hurter_id, float xpos, float ypos, float radius, float damage)
{
	//find all the objects that need to be hurt by us.
	World_objects::iterator i = objects->begin();
	while (i != objects->end()) {
  	if (i->second->hurtable()) {

			float xp=0;
			float yp=0;
			i->second->getpos(xp,yp);
			float distance = sqrt(SQR(xp-xpos)+SQR(yp-ypos));

			if (distance < radius*2) {
				float hitpoints;
				if (distance < radius)	
					hitpoints = damage;
				else
					hitpoints = damage*(1-(distance-radius) / radius);
 	
				Vector v;
				v.y = yp-ypos;
				v.normalize();
				v.scale(radius);
					
				i->second->hurt(hitpoints,v,hurter_id);
			}

		}
	
		i++;
	}
}


bool Serverworld::hit_object(float old_xpos, float old_ypos, float xpos, float ypos, int &cx, int &cy, int &id)
{
	int closest_objectid = -1;
	int closest_xc = -1;
	int closest_yc = -1;
	float closest_dist = 9999;
		
	World_objects::iterator i = objects->begin();
	while (i != objects->end()) {
		if (i->second->hurtable()) {
			float vx;
			float vy;
			i->second->getpos(vx,vy);
			long xhit = ROUND(xpos);
			long yhit = ROUND(ypos);

			//Then see if the path crosses the victim object..
			if (lines_intersect(
					ROUND(vx),ROUND(vy),ROUND(vx),ROUND(vy-8),
					ROUND(old_xpos),ROUND(old_ypos),ROUND(xpos),ROUND(ypos),
					&xhit,&yhit)){

				float thisdist = dist(xhit,yhit,old_xpos,old_ypos);

				if (thisdist<closest_dist) {
					closest_dist = thisdist;
					closest_objectid = i->second->getid();
					closest_xc = xhit;
					closest_yc = yhit;
				}
			}
		}
		i++;
	}

	//report the findings:		
	if (closest_objectid >= 0) {
		cx = closest_xc;
		cy = closest_yc;
		id = closest_objectid;
		return true;
	} else {
   	return false;
	}
}
