/***************************************************************************
                          tcp_server_connection.cpp  -  description
                             -------------------
    begin                : Sat Nov 13 1999
    copyright            : (C) 1999 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/

#include "tcp_server_connection.h"
#include "tcp_server.h"
#include "systemheaders.h"

//-------------------------------------------------------------------------
//This routine is called in the new thread:
void *start_server_connection(void *arg)
{
	Tcp_server_connection * conn = ((Tcp_server_connection *)arg);

	conn->execute();

	if (conn->socket > 0) {
  	shutdown(conn->socket,2);
		conn->socket = -1;
	}
	logmsg(lt_debug,"Thread ended");

 	conn->done = true; //mark the connection object for deletion

	return NULL;
}

//-------------------------------------------------------------------------
int Tcp_server_connection::accept(Tcp_server *tcp_server_, int server_socket)
{
	if (!done) return -1; //we cannot start a new thread on a connection that is not done.
	bedone = false;
	done = false;

	tcp_server = tcp_server_;
	unsigned int client_addr_size = sizeof(client_addr);
  socket = ::accept(server_socket, (sockaddr*)&client_addr, &client_addr_size);
	if (socket < 0) {
		done = true;
  	return -2;
	}

	//let the client join in this thread
	//this could be moved into the connection new thread to get rid of the blocking
	if (!joinclient()) {
		done = true;
  	return -4;
	}

	//If the server is multithreaded then create a thread, if not the server will poll us
	if (tcp_server->ismultithreaded()) {
	  int therad_error = pthread_create(&threaddata, NULL, start_server_connection, this);
		if (therad_error) {
			done = true;
  		return -3;
		}
	}

	return 0;
}

//-------------------------------------------------------------------------
//This handles the joining of a client:
bool Tcp_server_connection::joinclient()
{
	//start by configuring the socket:
	{
		int flag = 1;
  	int opt_error = setsockopt(socket, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));
		if (opt_error) {
			logmsg(lt_error,"Could not disable nagle on connection_socket");
			bedone = true;
			return false;
		}
	}

	Net_server_signature server_signature;
	if (send(&server_signature) < 0) {
		logmsg(lt_error,"Couldn't send server signature");
		bedone = true;
		return false;
	}
	logmsg(lt_debug,"signature sent");

	Net_joinrequest joinrequest;
	if (recv(&joinrequest) < 0) {
		logmsg(lt_error,"Couldn't recv joindata");
		bedone = true;
		return false;
	}
	logmsg(lt_debug,"Got joindata");
	
	Net_joinreply joinreply;
	char *map_data=NULL;
	tcp_server->serverworld->handle_join_request(
			joinrequest,//input, comes from the client
			player,  		//output, don't delete!
			joinreply,	//output
			map_data		//output, must free()
	);

  if (send(&joinreply) < 0) {
		logmsg(lt_error,"Couldn't send joinreply");
		bedone = true;
		return false;
	}

	logmsg(lt_debug,"sent joinreply");

  if (send(map_data, joinreply.compressed_size) < 0) {
		logmsg(lt_error,"Couldn't send map_data");
		bedone = true;
		return false;
	}

  free(map_data);
	logmsg(lt_debug,"sent map");

	int handshake=0;
	if (recv(&handshake,sizeof(handshake),15000) < 0) {
		logmsg(lt_error,"Couldn't recv post-map handshake");
		bedone = true;
		return false;
	}

	logmsg(lt_debug,"got post-map handshake");

	//Make sure that the serverworld sends out all objects in it's next run.
  tcp_server->serverworld->make_all_dirty();

	ack_update_id = current_update_id = tcp_server->update_queue.get_latest_update_id();

	return true;
}

//-------------------------------------------------------------------------
Tcp_server_connection::Tcp_server_connection()
{
	tcp_server = NULL;
	bedone = true;
	done = true;
	ack_update_id = -1;
	current_update_id = -1;
}

//-------------------------------------------------------------------------
Tcp_server_connection::~Tcp_server_connection()
{
//	player->bedone();

	//signal the thread that it ought to finish up now
	bedone = true;
	//wait for it to do so:
	while (!done) usleep(1000);
	logmsg(lt_debug,"connection terminated.");
}

//-------------------------------------------------------------------------
void Tcp_server_connection::poll()
{
	if (bedone) {
		if (!tcp_server->ismultithreaded())
			done = true;
		return;
	}

	fd_set readable;
	FD_ZERO(&readable);
	FD_SET(socket, &readable);

	fd_set writeable;
	FD_ZERO(&writeable);
	FD_SET(socket, &writeable);

	timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = tcp_server->ismultithreaded()?10000:2000;

	int retval = select(socket+1, &readable, &writeable, NULL, &tv);
	if (retval < 0) {
  	logmsg(lt_error,"select() failed in Tcp_server_connection::poll() : %i",retval);
		player->bedone();
		bedone = true;
    return;
	}

	//a set of controls are available, get it.
	if (FD_ISSET(socket,&readable)) {
		bool quit = false;
		int netstate = -1;
		recv(netstate);
		while (netstate != NETSTATE::DONE) {

      switch (netstate) {

        case NETSTATE::CONTROLS: {
		      if (recv(&player->controls) < 0) {
						logmsg(lt_error,"Couldn't recv(&player->controls) dropping client.");
						quit = true;
					}
					break;
				}

        case NETSTATE::UPDATEID: {
					recv(ack_update_id);
					break;
				}

        case NETSTATE::DONE: break;

        case NETSTATE::QUIT: {
					logmsg(lt_message,"Client politely requested to quit, going to let it.");
          quit = true;
					break;
				}

      	default: {
        	logmsg(lt_error,"Invalid netstate while recv()'in data from the client: %i",netstate);
					quit = true;
					return;
				}
			} //switch (netstate) {

			netstate = -1;
			recv(netstate);
			
			if (quit || netstate == -1) {
				player->bedone();
				logmsg(lt_debug,"Player has been told to remove itself from the world.");
				bedone = true;
				break;
			}
			
		}//while (netstate != NETSTATE::DONE) {

	}

	//the socket is ready for more data, let it have an update:
	if (FD_ISSET(socket,&writeable)) {
		int latest_update_id = tcp_server->update_queue.get_latest_update_id();

    while (current_update_id < latest_update_id) {
	    send(NETSTATE::UPDATEID);
	    send(current_update_id);
		
	    send(NETSTATE::WORLDUPDATE);
			Serializer_reader *reader = tcp_server->update_queue.get_update(current_update_id);
			if (!reader) {
				logmsg(lt_error,"Asked for a reader (%i) and I didn't get one! latest was: %i",current_update_id,latest_update_id);
				player->bedone();
				bedone = true;
				return;
			}
			if (send(reader) < 0) {
				logmsg(lt_error,"Couldn't send(&reader)");
				player->bedone();
				bedone = true;
				return;
			}

			current_update_id++;

    	send(NETSTATE::DONE);

		} //while (current_update_id < latest_update_id) {
	}	//if (FD_ISSET(socket,&writeable)) {
}

//-------------------------------------------------------------------------
//This is the routine that is run in the sepperate thread:
void Tcp_server_connection::execute()
{
	logmsg(lt_debug,"server connection thread started");

	//NOTE: This should not be done this way!
	//A much better way would be to create an object that that we can block while waiting for
	//(I just don't know how yet)
	while (!bedone) {
		poll();
		usleep(1000);
	}
	
	done = true;
}
